#
# tb.sed - trace branches
#
# tb adds tracing points in C code at each branch in the control flow.
# the traces generated can then be analysed to discover parts of code
# that get never executed during tests.

# tb will insert calls to the macro TB(n) at the following places 
# (indicated by _):
#   if ( ... ) _    while ( ... ) _   for ( ... ) _   else _
#   ? _ ... : _     && _      || _    case ... : _    default: _    
# n is an integer that is incremented after each TB(n) insertion.
# The actual text inserted is one of:
#   TB(n),  ( TB(n), ... )  TB(n);   { TB(n); ... } 
# macro lines and lines between /* TBOFF */ and /* TBON */ will be 
# ignored (treated as blank lines and passed verbatim)
#
# Limitations:
# - trigraphs not supported
# - deliberate syntax abuses not supported
# - preprocessor directive can confuse tb. For example code like:
#     #if A \ if(a) \ #endif \ b();        (\ standing for newline)
#   will be transformed into
#     #if A \ if(a) { TB(n); \ #endif \ b(); }
#   and similarly you should avoid code like:
#     #if A \ if(a) { #else \ if(b) { \ #endif
#     #if A \ if(a) a(); else \ #endif
#     ...
# - tb might add more {} and () than apparently necessary, to be able to
#   cover the following cases:
#     if(a) { int b; ... }   ->   if(a) { TB(n); { int b; ... } }
#     a && (int)b;  ->  a && ( TB(n), (int)b );
# - ||, && and ?: will be modified in constants too;
# - there was no attempt to optimize this for speed.

# e: in an expression (number of e is () depth); f between ? and : 
# a: statement (as many 'a's as statement depth);
# A: compound statement

/\/[/*] *TBOFF *\(\*\/\)\{0,1\}/,/\/[/*] *TBON *\(\*\/\)\{0,1\}/b

1{
  i\
#include "tb.h"
  # if the first line is a number, it is the initial value of n
  /^[0-9]\{1,\}$/{
    s//aA &/
    h
    d
  }
  # initialize n = 0 and state = a in hold buffer
  x
  s/^/aA 0/
  x
}

# hide ! behind `; add a space at the beginning and end (easier regexps)
s/`/`'/g
s/!/`-/g
s/@/`a/g
s/^/ /
s/$/ /

# insert '@' before assignment operators 
# =  *=  /=  %=  +=  -=  <<=  >>=  &=  ^=  |=
s/\([*/%+&^|-]=\)/@\1/g
s/<<=/@&/g
s/>>=/@&/g
s/\([^*/%+&^|<>-]\)=/\1@=/g
s/@*=@*=/==/g
# a++ =  (but note *a++ +=)
s/\([^+]+\)@+/\1+@/g
s/\([^-]-\)@-/\1-@/g
# /*=,  */=,  //=
s/\([*/]\)@/\1/g
# '='
s/'@='/'='/g
# !=
s/`@-=/`-=/g

# retrieve the current state

G
s/\(.*\)\n\([^ ]*\).*/!\2 \1/

# lexical states: d: in string; i: in comment; j: waiting to insert TB(n)
# 
# e: in an expression (number of e is () depth); f in ? ... : 
# a: statement (as many 'a's as statement depth);


# depending of the state, skip over lexical elements
:loop

# ignore preprocessor lines (lines starting with '#' and the following
# if ending with '\')

s/^!\([^j][^ ]* \) #/ #!M\1/
s/\(!M[^ ]* \)\(.*\\\)$/\2\1/
s/!M\([^ ]* \)\(.*[^\]\)$/\2!\1/

# skip space
s/\(![^kj][^ ]* \)\([[:space:]]\{1,\}\)/\2\1/

# every state except string and comments
# skip // comment
s/\(![^hkdij][^ ]* \)\(\/\/.*\)/\2\1/
# enter /* */ comment
s/!\([^hkdij][^ ]* \)\/\*/\/\*!i\1/
# skip / that do not begin a comment
s/\(![^Aahkdij][^ ]* \)\//\/\1/

# state i: /* */ comment
#   exit on */
s/!i\([^ ]* \)\([^*]*\*\/\)/\2!\1/
#   skip past any number of * not followed by a /
s/\(!i[^ ]* \)\([^*]*\**\)\([^/]\)/\2\1\3/
#   skip past any non-null number of non / followed by the end of line
s/\(!i[^ ]* \)\([^/]\{1,\}\)$/\2\1/

# state d: string
s/\(!d[^ ]* \)\([^"\]\{1,\}\)/\2\1/
s/\(!d[^ ]* \)\(\\.\)/\2\1/
s/\(!d[^ ]* \)\(\\\)$/\2\1/
s/!d\([^ ]* \)"/"!\1/


# Expressions
# Priority of operators: '&&',  '||',  '?' ':', assignment, ','. 
# All expressions end at a ';', ')' or ']'. expression states are:
# - e: a general expression 
# - E: expression ending at '?', ':', assignment, ','
# - F: expression ending at '||', '?', ':', assignment, ','
# - G: expression ending at '&&', '||', '?', ':', assignment, ','

#  skip ordinary stuff
s/\(![eEFG][^ ]* \)\([^]m()@:;,?&|"'/{}[]\{1,\}\)/\2\1/

#  skip 'x', '\x' and '\007'
s/\(![eEFG][^ ]* \)\('\\\{0,1\}.[^']*'\)/\2\1/

#  enter string
s/!\([eEFG][^ ]* \)"/"!d\1/

# if 'm', replace 'main' by tb_main
s/\([^_A-Za-z0-9]\)\(![eEFG][^ ]* \)m\(ain[^_A-Za-z0-9]\)/\1tb_m\2\3/
s/\(![eEFG][^ ]* \)m/m\1/

#  on ':', insert if underlying state is f (coming from ?)
s/!ef\([^ ]* \):/: (!j,E\1 /
s/!e\(b[^ ]* \):/!\1:/
s/!e\([^ ]* \):/:!a\1/

#  '([' -> one more depth; '])' -> decrease depth
s/!\([eEFG][^ ]* \)\([([]\)/\2!e\1/
s/!e\([^ ]* \)\([])]\)/\2!\1/

#  insert '(TB(n),' ... after '&&' and let state G add the ')'.
# note: we 
s/!\([eEF][^ ]* \)\(&&\)/\2(!j,G\1 /
#  skip a single &
s/\(![eEFG][^ ]* \)\(&[^&]\)/\2\1/

#  insert after '||'
s/!\([eE][^ ]* \)\(||\)/\2(!j,F\1 /
#  skip a single |
s/\(![eEFG][^ ]* \)\(|[^|]\)/\2\1/

#  insert after '?' and push state ef
s/!\([eEFG][^ ]* \)?/?!j,ef\1 /

# an expression-statement ends with ';'
s/!e\([^ ]* \);/;!\1/

# initialised element:  ... }
s/![eEFG]\([^ ]* \)}/!\1}/

# foo() {, switch() {, struct {
s/![eEFG]\([^ ]* \){/{!A\1/

# state EFG: insert ')' at the end of the current expression
s/!G\([^ ]* \)\(&&\)/)!\1\2/
s/![FG]\([^ ]* \)\(||\)/)!\1\2/
s/![EFG]\([^ ]* \)\([]?:,);@]\)/)!\1\2/
s/\(!e[^ ]* \),/,\1/

s/\(!e[^ ]* \)@/\1/



# statements


#
# identifiers
#

# 'case' -> skip an expression and push state b; 'default' -> push state b
s/!\(a[^ ]* \)case\([^_a-zA-Z0-9]\)/case!eb\1\2/
s/!\(a[^ ]* \)default\([^_a-zA-Z0-9]\)/default!b\1\2/

# 'do' -> catch matching 'while' with state g
s/!a\([^ ]* \)do\([^_a-zA-Z0-9]\)/do!ag\1\2/

# 'for' -> reach state h after skipping the (...)
s/!a\([^ ]* \)for\([^_a-zA-Z0-9]\)/for!caaeh\1\2/
# same for 'while' if not in a 'do' ... 'while'
s/!a\([^ ]* \)while\([^_a-zA-Z0-9]\)/while!ceh\1\2/
# same for 'if', with state l to look for 'else' after state h
s/!a\([^ ]* \)if\([^_a-zA-Z0-9]\)/if!cehl\1\2/

# state g: catch ending 'while' of 'do' ... 'while'
s/!g\([^ ]* \)while\([^_a-zA-Z0-9]\)/while!ce\1\2/

# state l: look for an optional 'else' after 'if'
s/!l\([^ ]* \)else\([^_a-zA-Z0-9]\)/else!h\1\2/
s/!l\([^ ]* \)\([^/[:space:]]\)/!\1\2/

# end of statement
s/!a\([^ ]* \)}/!\1}/
s/!a\([^ ]* \);/;!\1/

# anything else is treated as an expression: set state e
s/!a\([^ ]* \)\([^{};/[:space:]]\)/!e\1\2/

# begin compound statement
s/!a\([^ ]* \){/{!aA\1/

# anything but '}' in a compound statement is the start of another
# statement
s/!\(A[^ ]* \)\([^}/[:space:]]\)/!a\1\2/
s/!A\([^ ]* \)}/}!\1/

# state b: insert after ':'
s/!b\([^ ]* \):/:!j;\1/

# state c: look for '(' 
s/!c\([^ ]* \)(/(!\1/

# state h: insert '{TB();' and push state k. (this is needed even if a '{'
# is already there to allow declarations at the beginning of the compound.)
s/!h\([^ ]* \)/ {!j;ak\1 /

# state k: add a '}' after the end of statement
s/!k\([^ ]* \)/ }!\1/

# state j: replace with TB(n)
/!j/ {
  G
  s/!j\(.\)\(.*\)\n[^ ]* \(.*\)/ TB(\3)\1!\2/
  x
# increment n using Greg Ubben's trick
  s/$/+01234567891099990000999000990090/
  s/\(.\)\(9*\)+[^9]*\1\(.0*\).*\2\(0*\).*/\3\4/
  x
}

t loop

# quit reporting an error if the parser did not reach the end of line
/![^ ]* ./{
  s/^/#error /
  q
}

# save new state in hold buffer
H
x
s/[^ ]* \([0-9]*\)\n[^!]*!\([^ ]* \).*/\2\1/
# if at end, remove the state and write the number to tb.dat
$s/[^ ]* //w tb.dat
x
s/![^ ]* //

# 
s/@//g
s/`a/@/g

# remove space at end, unhide ! behind `
s/^ //
s/ $//
s/`-/!/g
s/`'/`/g


